#############################################################
#             Partial score and DIF functions               #
#############################################################
# Authors: Susu Zhang, Jitong Qi
# Source functions for process workshop illustration (Section 3: Applications)

library('Rcpp')
library('mirt')
# library('inline')
library('doSNOW')

# --- Base functions ------------------------------------------------
if(!requireNamespace("RcppArmadillo", quietly = TRUE)) { install.packages("RcppArmadillo") }

Rcpp::cppFunction(
  "arma::mat vec_matcross_cpp(const arma::vec& v1, const arma::mat& m1) {
  arma::mat cp = m1.t() * arma::diagmat(v1) * m1;
  return cp;
  }",
  depends = "RcppArmadillo"
)



Rcpp::cppFunction(
  "arma::mat inv_sympd_cpp(const arma::mat& m1) {
  return arma::inv(m1);
  }",
  depends = "RcppArmadillo"
)



Rcpp::cppFunction(
  "arma::mat matmult_cpp(const arma::mat& m1, const arma::mat& m2) {
  return m1 * m2;
  }",
  depends = "RcppArmadillo"
)



Rcpp::cppFunction(
  "arma::mat runi_cpp(int dim1, int dim2) {
  return arma::mat(dim1, dim2, arma::fill::randu);
  }",
  depends = "RcppArmadillo"
)










predict_glm <- function(model, newdata){
  as.vector(1/(1 + exp(-matmult_cpp(newdata, as.matrix(model$coefficients, ncol = 1))))) 
}

# simulate the L_2 distance test statistic under the null distribution
DIF_sim <- function(Xr, Xf, Yr, Yf){
  stopifnot(dim(Xr)[1] == length(Yr))
  stopifnot(dim(Xf)[1] == length(Yf))
  stopifnot(dim(Xr)[2] == dim(Xf)[2])
  
  ## Fit focal, reference, null models
  model_focal <- glm.fit(Xf, Yf, family = binomial(), intercept = F)
  model_ref <- glm.fit(Xr, Yr, family = binomial(), intercept = F)
  
  ## Calculate test vector z
  z <- predict_glm(model_focal, newdata = rbind(Xf, Xr)) - 
    predict_glm(model_ref, newdata = rbind(Xf, Xr))
  stopifnot(length(z) == (length(Yf) + length(Yr)))
  
  ## Calculate test statistic
  test_statistic <- sum(z^2)
  
  ## Return values
  return(test_statistic)
}


# get l2 distance and p value for a particular selection of process features
DIF_p_value_sim <- function(Xr, Xf, Yr, Yf, N = 1e2, sim = F){
  stopifnot(dim(Xr)[1] == length(Yr))
  stopifnot(dim(Xf)[1] == length(Yf))
  stopifnot(dim(Xr)[2] == dim(Xf)[2])
  Xr <- cbind(1, Xr)
  Xf <- cbind(1, Xf)
  
  nf <- length(Yf)
  nr <- length(Yr)
  ## Fit focal, reference, null models
  model_focal <- glm.fit(Xf, Yf, family = binomial(), intercept = F)
  model_ref <- glm.fit(Xr, Yr, family = binomial(), intercept = F)
  model_null <- glm.fit(rbind(Xf, Xr), c(Yf, Yr), family = binomial(), intercept = F)
  ## Get the predicted probability 
  focal_prob_f <- predict_glm(model_focal, newdata = Xf)
  ref_prob_r <- predict_glm(model_ref, newdata = Xr)
  ## Calculate covariance matrix of estimators
  XftWfXf <- vec_matcross_cpp(focal_prob_f*(1-focal_prob_f), Xf)
  XrtWrXr <- vec_matcross_cpp(ref_prob_r*(1-ref_prob_r), Xr)
  
  Cov_f <- inv_sympd_cpp(XftWfXf)
  Cov_r <- inv_sympd_cpp(XrtWrXr)
  
  ## z test statistics of last added feature 
  K <- dim(Xr)[2]
  z_f <- model_focal$coefficients[K]/sqrt(Cov_f[K,K])
  z_r <- model_ref$coefficients[K]/sqrt(Cov_r[K,K])
  
  if((abs(z_f) < qnorm(0.975))|(abs(z_r) < qnorm(0.975))){
    if(K > 2){
      stop('In-significant')
    }
  }
  
  ## z test statistics of theta 
  z_f <- model_focal$coefficients[2]/sqrt(Cov_f[2,2])
  z_r <- model_ref$coefficients[2]/sqrt(Cov_r[2,2])
  
  if((abs(z_f) < qnorm(0.975))|(abs(z_r) < qnorm(0.975))){
    if(K > 2){
      stop('In-significant')
    }
  }
  
  if(sim){
    ## Calculate weights matrices
    
    null_prob_f <- predict_glm(model_null, newdata = Xf)
    stopifnot(length(null_prob_f) == length(Yf))
    
    null_prob_r <- predict_glm(model_null, newdata = Xr)
    stopifnot(length(null_prob_r) == length(Yr))
  }
  
  ## Calculate test vector z
  z <- predict_glm(model_focal, newdata = rbind(Xf, Xr)) - 
    predict_glm(model_ref, newdata = rbind(Xf, Xr))
  stopifnot(length(z) == (length(Yf) + length(Yr)))
  
  ## Calculate test statistic
  test_statistic <- sum(z^2)
  
  if(sim){
    ## Simulate N times
    test_statistic_samples <- foreach(k = 1:N, .combine = c, .noexport = c('matmult_cpp', 'runi_cpp')) %dopar% {
      Yf_sim <- as.numeric(runi_cpp(nf, 1) < null_prob_f)
      Yr_sim <- as.numeric(runi_cpp(nr, 1) < null_prob_r)
      DIF_sim(Xr, Xf, Yr_sim, Yf_sim)
    }
    
    ## Calculate p-value
    p_value <- mean(test_statistic_samples > test_statistic)
  }else{
    p_value = -1
  }
  
  ## Return values
  return(c('L2' = sqrt(test_statistic/length(z)), 'p-value' = p_value))
}


# --- Partial score functions -----------------------------------------------

# fit mirt model (item parameters) on set of items
get_model_mirt <- function(scores,          # matrix of item responses (colnames are item names)
                           types,           # named vector of item type (e.g., 2PL, GRM)
                           itemset){        # set of items used for parameter calibration
  model <- mirt(data = scores[,itemset], model = 1, itemtype = types[itemset],verbose = F)
  return(model)
}

# get theta based on subset of item responses with known irt pars
get_traits_mirt<-function(scores,           # matrix of item responses (scores[,itemset] in get_model_mirt)
                          model,            # output of get_model_mirt
                          itemset,          # subset of item responses used for ability estimation
                          method = 'EAP'){  # ability estimation method
  scores[,which(! colnames(scores) %in% itemset)] <- NA
  res <- fscores(model, response.pattern = as.matrix(scores), method = method, full.scores.SE = TRUE)
  theta <- res[,'F1']
  if(method == 'ML'){
    theta[which(theta == Inf)] <- max(theta[which(theta<Inf)])+.5
    theta[which(theta == -Inf)] <- min(theta[which(theta>-Inf)])-.5
  }
  return(theta)
}


# --- DIF functions -------------------------------------------------

# Foward selection based on L2 distance 

DIF_forward <- function(features,          # process features from DIF item
                        response,          # response vector on the DIF item 
                        group,             # vector of group membership, 'F' or 'R'
                        theta_ref,         # latent trait estimate on the anchor set (items w/o DIF)
                        K = NULL           # max number of feature dimensions to add (default to feature dimensions)
){
  # reorganize data
  Xs <- cbind(theta_ref, features)
  Xf <- Xs[which(group == 'F'),]
  Xr <- Xs[which(group == 'R'),]
  Yf <- response[which(group == 'F')]
  Yr <- response[which(group == 'R')]
  
  if( is.null(K)) K <- ncol(features)
  results_by_k <- matrix(NA, (K+1), 2)
  colnames(results_by_k) <- c('L_2', 'p')

  # initialize --- theta only in irt model
  
  Xf_tmp <- as.matrix(Xf[,1])
  Xr_tmp <- as.matrix(Xr[,1])
  results_by_k[1,] <- DIF_p_value_sim(Xf_tmp, Xr_tmp, Yf, Yr, sim = T)
  cat('---   Initial model    ------ \n  L2 dist: ',results_by_k[1, 1],
      '\n  p-value: ',results_by_k[1,2],'\n')
  
  # add process features one by one, each feature chosen to minimize l2 distance
  add_index <- c()
  for(dim in 1:K){
    remain_index <- setdiff((1:ncol(features)), add_index)
    
    # calculate l_2 for each added candidate feature
    p_value_temp <- foreach(k = 1:length(remain_index), .combine = c, .noexport = c('matmult_cpp', 'runi_cpp', 'inv_sympd_cpp', 'vec_matcross_cpp')) %dopar% {
      # already selected
      Xs_tmp <- as.matrix(Xs[,c(1, (add_index+1))])
      # add new candidate feature
      Xs_tmp <- cbind(Xs_tmp, Xs[,(remain_index[k]+1)])
      # split to focal and reference
      X_focal <- Xs_tmp[which(group == 'F'),]
      X_ref <- Xs_tmp[which(group == 'R'),]
        tryCatch(
        DIF_p_value_sim(Xf = X_focal, Xr = X_ref, Yf = Yf, Yr = Yr, sim = F),
        error = function(e) c(1, 0)
      )
    }
    p_value_temp <- matrix(p_value_temp, nrow = 2)
    
    stopifnot(dim(p_value_temp) == c(2, length(remain_index)))
    if(sum(p_value_temp[1,]) == length(remain_index)){
      break
    }
    # select feature with lowest L_2
    k <- which.min(p_value_temp[1,])
    add_index <- c(add_index, remain_index[k])
    
    # get p value
    Xs_tmp <- Xs[,c(1,(add_index+1))]
    # split to focal and reference
    X_focal <- Xs_tmp[which(group == 'F'),]
    X_ref <- Xs_tmp[which(group == 'R'),]
    results_by_k[dim+1,] <- tryCatch(
      DIF_p_value_sim(Xf = X_focal, Xr = X_ref, Yf = Yf, Yr = Yr, sim = T),
      error = function(e) c(1, 0, 0, 0, 0, 0)
    )
    
    
    cat('--- K =',dim,'Adding ft', remain_index[k],'------ \n  L2 dist: ',results_by_k[(dim+1), 1],
        '\n  p-value: ',results_by_k[(dim+1),2],'\n')
    
  }
  
  results_by_k <- cbind(results_by_k, c(NA, add_index))
  colnames(results_by_k)[3] <- 'Added_ft'
  return(results_by_k)
  
}


# Updated IRT model w/ Nuisance 
get_itempar_DIF <- function(features,            # process features from DIF item
                          response,            # response vector on the DIF item
                          theta_ref,           # latent trait estimate on anchor set (items w/o DIF)
                          DIF_forward_result,  # output matrix from DIF_forward (index of selected features, l2, p)
                          K = NULL             # feature dimension to add (default to dimension with smallest L2)  
){
  # Default K: lowest L2 distance between focal and reference irfs
  if(is.null(K)){
    K <- which.min(DIF_forward_result[,'L_2'])-1
  }
  
  # get feature set
  add_index <- DIF_forward_result[1:K+1,'Added_ft']
  
  # fit logistic
  Xs <- cbind(1, theta_ref, features[,add_index])
  pars <- glm.fit(Xs, response, family = binomial(), intercept = F)$coefficients
  d <- pars[1]
  a <- pars[2]
  omegas <- numeric(ncol(features))
  omegas[add_index] <- pars[-(1:2)]/a
  
  return(list(a = a, d = d, omegas = omegas))
}


# DIF model item response function (w/ nuisance eta)
p_dif <- function(theta, eta,
                    a, d){
  return(1/(1+exp(-(a*(theta+eta)+d))))
}

# loglikelihood function with DIF irf w/ nuisance eta 
ll_dif <- function(theta, etas, as, ds, response){ # theta: scalar, etas/as/ds/response: vec of length n_item for individual i
  logit <- as*(theta+etas)+ds
  loglikelihood <- sum(logit[response == 1]) - sum(log(1 + exp(logit)))
  loglikelihood <- -loglikelihood/length(response)
  return(loglikelihood)
}


# Re-estimate theta with new IRFs for DIF items 
update_theta <- function(responses,             # response matrix
                         item_DIF,              # vector of DIF item index
                         fts_DIF,               # list of DIF item features, each element is ft mat of an item
                         as_anchor,             # vector of a parameters for items w/o DIF
                         ds_anchor,             # vector of d parameters for items w/o DIF
                         itempars_DIF,          # list of DIF item parameters, each element is output of get_itempar_DIF
                         method = 'EAP'         # trait estimation method; 'EAP': N(0,1) prior; 'MLE': bounds are -4, 4.
){
  
  # get nuisance of each item (for no dif items this is 0)
  etamat <- matrix(0, nrow(responses), ncol(responses))
  for(i in 1:length(item_DIF)){
    fts <- fts_DIF[[i]]
    itempars <- itempars_DIF[[i]]
    etamat[,item_DIF[i]] <- fts %*% itempars$omegas
  }
  
  # get item slope and intercepts
  as <- ds <- rep(NA, ncol(responses))
  as[-item_DIF] <- as_anchor
  ds[-item_DIF] <- ds_anchor
  as[item_DIF] <- sapply(itempars_DIF, function(x) x$a)
  ds[item_DIF] <- sapply(itempars_DIF, function(x) x$d)
  
  # estimate theta
  if(method == 'EAP'){
    theta_pts <- seq(-4, 4, .2)
    theta_density <- rep(1, nrow(responses)) %o% (dnorm(theta_pts) * .2) 
    like_theta <- matrix(NA, nrow(responses), length(theta_pts))
    for(t in 1:length(theta_pts)){
      theta <- theta_pts[t]
      
      # get likelihood | theta
      like <- rep(1, nrow(responses))
      for(i in 1:ncol(responses)){
        like <- like * (p_dif(theta, etamat[,i], as[i], ds[i])^responses[,i]) * 
          ((1-p_dif(theta, etamat[,i], as[i], ds[i]))^(1-responses[,i]))
      }
      like_theta[,t] <- like
    }
    
    theta_est <- colSums(theta_pts * t(theta_density * like_theta))/(rowSums(theta_density * like_theta))
    
  }
  if(method == 'MLE'){
    theta_est <- rep(NA, nrow(responses))
    for(n in 1:nrow(responses)){
      theta_est[n] <- optimise(ll_dif, lower = -4, upper = 4, as = as, ds = ds, 
                               etas = etamat[n,], response = responses[n,])$minimum
    }
  }
  return(theta_est)
}


